"""
Utility functions for doing low-level, general-purpose EBML reading and writing.
"""

from io import IOBase
import logging
import os.path
from pathlib import Path

from ebmlite import loadSchema

from .importer import filterTime, openFile, _getSize
from .dataset import Dataset

# ==============================================================================
#
# ==============================================================================

logger = logging.getLogger('idelib')


# ==============================================================================
#
# ==============================================================================

def verify(data, schema=None):
    """ Basic sanity-check of data validity. If the data is bad an exception
        will be raised. The specific exception varies depending on the problem
        in the data.

        :keyword schema: The full module name of the EBML schema.
        :return: `True`. Any problems will raise exceptions.
    """
    if schema is None:
        schema = loadSchema('mide_ide.xml')
    return schema.verify(data)


# ==============================================================================
#
# ==============================================================================

def extractTime(doc, out, startTime=0, endTime=None, channels=None,
                updater=None):
    """ Efficiently extract data within a certain interval from an IDE file.
        Note that due to the way data is stored in an IDE, the exported
        interval will be slightly wider than the specified start and end
        times; this ensures the data is copied verbatim and without loss.

        :param doc: A loaded `Dataset` or the name of an IDE file.
        :param out: A filename or stream to which to save the extracted data.
        :param startTime: The start of the extraction range, relative to the
            recording's start.
        :param endTime: The end of the extraction range, relative to the
            recording's end.
        :param channels: A list of channel IDs to specifically export. If
            `None`, all channels will be exported. Note excluded channels will
            still appear in the new IDE's `channels` dictionary, but the file
            will contain no data for them.
        :param updater: A function (or function-like object) to notify as
            work is done. It should take four keyword arguments: `count` (the
            current line number), `total` (the total number of samples), `error`
            (an unexpected exception, if raised during the import), and `done`
            (will be `True` when the split is complete). If the updater object
            has a `cancelled` attribute that is `True`, the import will be
            aborted. The default callback is `None` (nothing will be notified).
        :return: The total number of bytes written, and total number of
            ChannelDataBlock elements copied.
    """
    if isinstance(doc, (str, Path)):
        doc = openFile(doc)

    if endTime is None:
        endTime = float('infinity')

    copiedBytes = 0

    # Updater stuff
    totalSize = 1
    increment = 50
    if updater:
        totalSize = _getSize(doc.ebmldoc)
        # FUTURE: Set `increment` based on `totalSize`?

    if isinstance(out, (str, Path)):
        fs = open(out, 'wb')
    elif isinstance(out, IOBase) or hasattr(out, 'seek'):
        fs = out
    else:
        raise TypeError("unsupported type for output; expected filename "
                        "or stream, got {} ".format(type(out)))

    try:
        for n, el in enumerate(filterTime(doc, startTime, endTime, channels=channels), 1):
            if updater:
                if updater.cancelled:
                    break

                if n % increment == 0:
                    updater(percent=el.offset/totalSize)

            if el is not None:
                data = el.getRaw()
                fs.write(data)
                copiedBytes += len(data)

    except (StopIteration, KeyboardInterrupt):
        pass

    finally:
        if isinstance(out, (str, Path)):
            fs.close()

    if updater:
        updater(done=True)

    return copiedBytes


# ==============================================================================
#
# ==============================================================================

CHUNK_SIZE = 512 * 1024  # Size of chunks when looking for last sync
SYNC = b'\xfa\x84ZZZZ'  # The raw EBML of a Sync element


def _getLastSync(stream, length=None):
    """
    Retrieve the offset of the last `Sync` element with data after it in
    an IDE.

    :param stream: A file stream or file-like object (must implement `tell()`
        and `seek()`).
    :param length: The length of the file, if already known. The length
        will be calculated if `None`.
    :return: The offset of the last EBML `Sync` element in the file
    """

    if not (hasattr(stream, 'seek') and hasattr(stream, 'tell')):
        raise TypeError("IDE file had bad stream type ({})".format(type(stream)))

    originalPos = stream.tell()
    offset = length or _getSize(stream)

    while True:
        offset = max(0, offset - CHUNK_SIZE)
        stream.seek(offset)
        chunk = stream.read(CHUNK_SIZE + len(SYNC))
        sync_idx = chunk.find(SYNC)
        if sync_idx > -1:
            # Make sure there's data after it by checking for ChannelDataBlock
            # IDs. This is somewhat brittle, since the IDs are 1 byte, but it
            # is not likely for anything but a ChannelDataBlock would have the
            # IDs for ChannelDataBlock and ChannelIDRef in order. Even if
            # they are in the payload, it still implies a ChannelDataBlock.
            block_idx = chunk.find(b'\xa1', sync_idx)
            if -1 < block_idx < chunk.rfind(b'\xb0', block_idx):
                offset += sync_idx
                break
        if offset <= 0:
            break

    stream.seek(originalPos)
    return max(offset, 0)


def _getBlockTime(doc, el):
    """
    Internal utility function to quickly get the ID and start and end times
    of an IDE `ChannelDataBlock`. Works directly with the EBML.

    :param doc: The opened IDE `idelib.dataset.Dataset`
    :param el: The `ChannelDataBlock EBML element to process
    :return: A tuple containing channel ID, start time, and end time.
    """
    start = end = None
    chId = blockStart = blockEnd = None
    for subEl in el:
        if subEl.name == "ChannelIDRef":
            chId = subEl.value
        elif subEl.name == "StartTimeCodeAbs":
            blockStart = subEl.value
        elif subEl.name == "EndTimeCodeAbs":
            blockEnd = subEl.value

    # TODO: Modulus correction, if still needed (i.e., very old files)
    blockEnd = blockEnd or blockStart
    scalar = doc._parsers['ChannelDataBlock'].timeScalars.get(chId, 1)
    if blockStart is not None:
        # logger.warning("getLength: {} missing <EndTimeCodeAbs> subelement, skipping.".format(el))
        start = blockStart // (1.0 / scalar)
    if blockEnd is not None:
        # logger.warning("getLength: {} missing <EndTimeCodeAbs> subelement, skipping.".format(el))
        end = blockEnd // (1.0 / scalar)

    return chId, start, end


def _getLength(doc):
    """
    Retrieve the start and end times of an `idelib.dataset.Dataset` (e.g. an
    open/imported IDE file). Used by `getLength()`.

    :param doc: The opened IDE file
    :return: The timestamps of the first and last samples in the file, in
        microseconds, relative to the start of the recording.
    """
    start = float('infinity')
    end = 0

    # Get starting time, reading several blocks since different channels
    # may not be in chronological order.
    count = 0
    for el in doc.ebmldoc:
        if el.name != "ChannelDataBlock":
            continue
        _chId, blockStart, _blockEnd = _getBlockTime(doc, el)
        start = min(blockStart, start) if blockStart is not None else blockStart

        count += 1
        if count > 10:  # Note: number is somewhat arbitrary
            break

    # Get ending time by jumping to a Sync near the end and reading blocks
    # from there to the end of file.
    stream = doc.ebmldoc.stream

    # Use the cached EBML document size if the IDE was already fully imported,
    # otherwise use `None` so `_getLastSync()` will calculate it (faster than
    # getting uncached `doc.ebmldoc.size`)
    streamLength = None if doc.loading else doc.ebmldoc.size
    stream.seek(_getLastSync(stream, streamLength))
    temp_ebml = doc.ebmldoc.schema.load(stream)

    for el in temp_ebml:
        if el.name != "ChannelDataBlock":
            continue

        _chId, _blockStart, blockEnd = _getBlockTime(doc, el)
        if blockEnd:
            end = max(blockEnd, end)

    return start, end


def getLength(doc):
    """
    Efficiently retrieve the start and end times of an IDE file, without it
    having to be fully imported.

    :param doc: The IDE filename, an `idelib.dataset.Dataset`, or stream
        containing IDE data.
    :return: The timestamps of the first and last samples in the file, in
        microseconds, relative to the start of the recording.
    """
    # This really just wraps `_getLength()` in order to handle different
    # source `doc` types
    if isinstance(doc, str):
        with open(doc, 'rb') as f:
            return _getLength(openFile(f))
    elif isinstance(doc, Dataset):
        return _getLength(doc)
    elif isinstance(doc, IOBase):
        return _getLength(openFile(doc))

    raise TypeError("getLength() needs a filename, stream, or Dataset; got {}".format(type(doc)))


# ==============================================================================
#
# ==============================================================================


def getExitCondition(recording):
    """ Get the ``ExitCond`` Attribute from the end of a recording, if present.
        The result will be an integer:

        * 1: Button press
        * 2: USB connection
        * 3: Recording time limit reached
        * 4: Low battery
        * 5: File size limit reached
        * 128: I/O error (can occur if disk is full or 4GB FAT32 size limit
          reached.

    :param recording: The IDE filename, an `idelib.dataset.Dataset`, an
        `ebmlite.core.Dataset`, or stream containing IDE data.
    """
    result = None

    if hasattr(recording, 'filename'):
        # A `Dataset` or `ebmlite.Document`
        recording = recording.filename

    if isinstance(recording, str):
        with open(recording, "rb") as fs:
            return getExitCondition(fs)

    if not (hasattr(recording, 'seek') and hasattr(recording, 'tell')):
        raise TypeError("IDE file had bad stream type ({})".format(type(recording)))

    offset = recording.tell()

    recording.seek(_getSize(recording) - CHUNK_SIZE)
    data = recording.read()
    try:
        # Seek out the exit condition Attribute by the
        # string of its name, then offset to where the
        # data is expected to be.
        idx = data.index(b"ExitCond") + 11
        if idx <= len(data):
            result = data[idx]
    except (IOError, IndexError, ValueError) as e:
        logger.warning(e)

    recording.seek(offset)
    return result



"""

"""

from collections import Counter
from datetime import datetime
import os.path
import sys
from time import time as time_time
from time import sleep
import warnings

import struct

from . import transforms
from .dataset import Dataset
from . import parsers


#===============================================================================
# 
#===============================================================================

# from dataset import __DEBUG__

import logging
logger = logging.getLogger('idelib')
logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s")


#===============================================================================
# Defaults
#===============================================================================

# from parsers import AccelerometerParser

# Legacy hard-coded sensor/channel mapping. Used when importing files recorded 
# on SSX running old firmware, which does not contain self-description data.
DEFAULTS = {
    "sensors": {
        0x00: {"name": "832M1 Accelerometer"},
        0x01: {"name": "MPL3115 Temperature/Pressure"}
    },
    
    "channels": {
            0x00: {"name": "Accelerometer XYZ",
    #                 "parser": struct.Struct("<HHH"), 
    #                 "transform": 0, #calibration.AccelTransform(),
                    "parser": struct.Struct("<HHH"),
                    "transform": 0,
                    "subchannels":{0: {"name": "Accelerometer Z", 
                                       "axisName": "Z",
                                       "units":('Acceleration','g'),
                                       "displayRange": (-100.0,100.0),
                                       "transform": 3,
                                       "warningId": [0],
                                       "sensorId": 0,
                                     },
                                  1: {"name": "Accelerometer Y", 
                                      "axisName": "Y",
                                      "units":('Acceleration','g'),
                                      "displayRange": (-100.0,100.0),
                                      "transform": 2,
                                      "warningId": [0],
                                      "sensorId": 0,
                                      },
                                  2: {"name": "Accelerometer X", 
                                      "axisName": "X",
                                      "units":('Acceleration','g'),
                                      "displayRange": (-100.0,100.0),
                                      "transform": 1,
                                      "warningId": [0],
                                      "sensorId": 0,
                                      },
                                },
                   },
            0x01: {"name": "Pressure/Temperature",
                   "parser": parsers.MPL3115PressureTempParser(),
                   "subchannels": {0: {"name": "Pressure", 
                                       "units":('Pressure','Pa'),
                                       "displayRange": (0.0,120000.0),
                                      "sensorId": 1,
                                       },
                                   1: {"name": "Temperature", 
                                       "units":('Temperature','\xb0C'),
                                       "displayRange": (-40.0,80.0),
                                      "sensorId": 1,
                                       }
                                   },
                   "cache": True,
                   "singleSample": True,
                   },
    },
    
    "warnings": [{"warningId": 0,
                   "channelId": 1,
                   "subchannelId": 1,
                   "low": -20.0,
                   "high": 60.0
                   }]
}


def createDefaultSensors(doc, defaults=None):
    """ Given a nested set of dictionaries containing the definition of one or
        more sensors, instantiate those sensors and add them to the dataset
        document.
    """
#     logger.info("creating default sensors")
    defaults = defaults or DEFAULTS
    sensors = defaults['sensors'].copy()
    channels = defaults['channels'].copy()
    warnings = defaults['warnings']
    
    if doc.recorderInfo:
        # TODO: Move device-specific stuff out of the main importer
        rtype = doc.recorderInfo.get('RecorderTypeUID', 0x10)
        if rtype | 0xff == 0xff:
            # SSX recorders have UIDs that are zero except the least byte.
            SSX_ACCEL_RANGES = {
               0x10: (-25,25),
               0x12: (-100,100),
               0x13: (-200,200),
               0x14: (-500, 500),
               0x15: (-2000, 2000),
               0x16: (-6000, 6000)
            }
            rrange = SSX_ACCEL_RANGES.get(rtype & 0xff, (-25,25))
            transform = transforms.AccelTransform(*rrange)
            ch0 = channels[0x00]
            ch0['transform'] = transform
            for i in range(3):
                ch0['subchannels'][i]['displayRange'] = rrange

    for sensorId, sensorInfo in sensors.items():
        doc.addSensor(sensorId, sensorInfo.get("name", None))
        
    for chId, chInfo in channels.items():
        chArgs = chInfo.copy()
#         chArgs['sensor'] = sensor
        subchannels = chArgs.pop('subchannels', None)
        channel = doc.addChannel(chId, **chArgs)
        if subchannels is None:
            continue
        for subChId, subChInfo in subchannels.items():
            channel.addSubChannel(subChId, **subChInfo)
    
    for warn in warnings:
        doc.addWarning(**warn)
    

#===============================================================================
# Parsers/Element Handlers
#===============================================================================

# Parser importer. These are taken from the module by type. We may want to 
# create the list of parser types 'manually' prior to release; it's marginally 
# safer.
ELEMENT_PARSER_TYPES = parsers.getElementHandlers()


def instantiateParsers(doc, parserTypes=None):
    """ Create a dictionary of element parser objects keyed by the name of the
        element they handle. Handlers that handle multiple elements have
        individual keys for each element name.
    """
    parserTypes = parserTypes or ELEMENT_PARSER_TYPES
    elementParsers = {}
    for t in parserTypes:
        p = t(doc)
        if isinstance(t.elementName, str):
            elementParsers[t.elementName] = p
        else:
            for name in t.elementName:
                elementParsers[name] = p
    return elementParsers


#===============================================================================
# Updater callbacks
#===============================================================================

class SimpleUpdater(object):
    """ A simple text-based progress updater.
        :ivar cancelled: If set to `True`, the job using the updater will abort. 
        :ivar paused: If set to `True`, the job using the updater will pause.
    """
    
    def __init__(self, cancelAt=1.0, quiet=False):
        """ Constructor.
            :keyword cancelAt: A percentage at which to abort the import. For
                testing purposes.
        """
        self.cancelled = False
        self.paused = False
        self.startTime = None
        self.cancelAt = cancelAt
        self.estSum = None
        self.quiet = quiet
    
    def dump(self, s):
        if not self.quiet:
            sys.stdout.write(s)
    
    def __call__(self, count=0, total=None, percent=None, error=None, 
                 starting=False, done=False):
        if percent >= self.cancelAt:
            self.cancelled=True
        if self.startTime is None:
            self.startTime = datetime.now()
        if starting:
            logger.info("Import started at %s" % self.startTime)
            return
        if done:
            logger.info("Import completed in %s" % (datetime.now() - self.startTime))
            logger.info("Original estimate was %s" % self.estSum)
        else:
            self.dump('\x0d%s samples read' % count)
            if percent is not None:
                p = int(percent*100)
                self.dump(' (%d%%)' % p)
                if p > 0 and p < 100:
                    d = ((datetime.now() - self.startTime) / p) * (100-p)
                    self.dump(' - est. completion in %s' % d)
                    if self.estSum is None:
                        self.estSum = d
                else:
                    self.dump(' '*25)
            sys.stdout.flush()
    

#===============================================================================
#
#===============================================================================

def _getSize(stream, chunkSize=512 * 1024):
    """
    Get the length of a stream from its data.

    :param stream: A file stream or file-like object (must implement `tell()`
        and `seek()`).
    :returns: The total length of the file.
    """
    if not (hasattr(stream, 'seek') and hasattr(stream, 'tell')):
        raise TypeError('Cannot get size of non-stream {}'.format(type(stream)))

    # If it's a real file, no problem!
    if hasattr(stream, 'name'):
        if os.path.isfile(stream.name):
            return os.path.getsize(stream.name)

    originalPos = stream.tell()

    # Grab chunks until less is read than requested.
    thisRead = chunkSize
    while thisRead == chunkSize:
        thisRead = len(stream.read(chunkSize))

    eof = stream.tell()
    stream.seek(originalPos)
    return eof


#===============================================================================
# ACTUAL FILE READING HAPPENS BELOW
#===============================================================================

def importFile(filename='', startTime=None, endTime=None, channels=None,
               updater=None, parserTypes=None, defaults=None, name=None,
               quiet=False, **kwargs):
    """ Create a new Dataset object and import the data from a MIDE file. 
        Primarily for testing purposes. The GUI does the file creation and 
        data loading in two discrete steps, as it will need a reference to 
        the new document before the loading starts.
        :see: `readData()`
    """
    # FUTURE: Remove `kwargs` and this conditional warning.
    if kwargs:
        warnings.warn(DeprecationWarning(
                'Some importFile() updater-related arguments have been deprecated. '
                'Ignored arguments: {}'.format(', '.join(repr(k) for k in kwargs))))

    defaults = defaults or DEFAULTS

    stream = open(filename, "rb")
    doc = openFile(stream, updater=updater, name=name, parserTypes=parserTypes,
                   defaults=defaults, quiet=quiet)
    readData(doc, startTime=startTime, endTime=endTime, channels=channels,
             updater=updater, parserTypes=parserTypes, defaults=defaults)
    return doc


def openFile(stream, updater=None, parserTypes=None, defaults=None, name=None,
             quiet=False):
    """ Create a `Dataset` instance and read the header data (i.e. non-sample-
        data). When called by a GUI, this function should be considered 'modal,' 
        in that it shouldn't run in a background thread, unlike `readData()`. 
        
        :param stream: The file or file-like object containing the EBML data.
        :param updater: A function (or function-like object) to notify as
            work is done. It should take four keyword arguments: `count`
            (the current line number), `total` (the total number of samples),
            `error` (an unexpected exception, if raised during the import),
            and `done` (will be `True` when the export is complete). If the
            updater object has a `cancelled` attribute that is `True`, the
            import will be aborted. The default callback is `None` (nothing
            will be notified).
        :param parserTypes: A collection of `parsers.ElementHandler` classes.
        :param defaults: A nested dictionary containing a default set of
            sensors, channels, and subchannels. These will only be used if
            the dataset contains no sensor/channel/subchannel definitions. 
        :param name: An optional name for the Dataset. Defaults to the
            base name of the file (if applicable).
        :param quiet: If `True`, non-fatal errors (e.g. schema/file version
            mismatches) are suppressed.
        :return: The opened (but still 'empty') `dataset.Dataset`
    """
    defaults = defaults or DEFAULTS
    parserTypes = parserTypes or ELEMENT_PARSER_TYPES

    if isinstance(stream, str):
        stream = open(stream, 'rb')
    
    doc = Dataset(stream, name=name, quiet=quiet)
    doc.addSession()

    if doc._parsers is None:
        doc._parsers = instantiateParsers(doc, parserTypes)
    
    elementParsers = doc._parsers
    
    try:
        for r in doc.ebmldoc:
            if getattr(updater, "cancelled", False):
                doc.loadCancelled = True
                break
            if r.name not in elementParsers:
                continue
            parser = elementParsers[r.name]
            if parser.makesData():
                break
            parser.parse(r) 
            
    except IOError as e:
        if e.errno is None:
            # The EBML library raises an empty IOError if it hits EOF.
            # TODO: Handle other cases of empty IOError (lots in python-ebml)
            doc.fileDamaged = True
        elif updater:
            updater(error=e)

    except TypeError as e:
        logger.exception(e)
        # This can occur if there is a bad element in the data
        # (typically the last)
        doc.fileDamaged = True

    if not doc.sensors:
        # Got data before the recording props; use defaults.
        if defaults is not None:
            createDefaultSensors(doc, defaults)
            
    doc.updateTransforms()
    return doc


def filterTime(doc, startTime=0, endTime=None, channels=None):
    """ Efficiently read data within a certain interval from an IDE file.
        Note that due to the way data is stored in an IDE, the exported
        interval will be slightly wider than the specified start and end
        times; this ensures the data is copied verbatim and without loss.

        :param doc: An opened (but not yet fully imported) `Dataset`.
        :param startTime: The start of the extraction range, relative to the
            recording's start.
        :param endTime: The end of the extraction range, relative to the
            recording's end.
        :param channels: A list of channel IDs to process. If `None` (the
            default), all channels are processed.
        :yields: Elements from the `Dataset`'s EBML file, excluding
            `ChannelDataBlock`s outside of the specified time range, or
            `None`.
    """
    if startTime == endTime and startTime is not None:
        raise ValueError('startTime and endTime must differ')

    startTime, endTime = sorted((startTime or 0,
                                 endTime or float('infinity')))

    # Dictionaries (and similar) for tracking progress of each ChannelDataBlock
    # element handled. All keyed by channel ID (ChannelIDRef).
    lastBlocks = {}  # Previous element w/ its start and end times
    channelsWritten = Counter()  # Number of elements extracted per channel
    finished = {}  # Channels that have completed extraction

    try:
        timeScalars = doc._parsers['ChannelDataBlock'].timeScalars

        for el in doc.ebmldoc:
            if el.name == 'ChannelDataBlock':
                # Get ChannelIDRef, StartTimeCodeAbs, and EndTimeCodeAbs;
                # usually the 1st three, in order, but don't assume!
                chId = blockStart = blockEnd = None
                for subEl in el:
                    if subEl.name == "ChannelIDRef":
                        chId = subEl.value
                    elif subEl.name == "StartTimeCodeAbs":
                        blockStart = subEl.value
                    elif subEl.name == "EndTimeCodeAbs":
                        blockEnd = subEl.value

                blockEnd = blockEnd or blockStart
                if chId is None:
                    logger.warning("Extractor: {} missing <ChannelIDRef> subelement, skipping.".format(el))
                    continue
                if blockStart is None:
                    logger.warning("Extractor: {} missing <StartTimeCodeAbs> subelement, skipping.".format(el))
                    continue

                if finished.setdefault(chId, False):
                    yield None
                    continue

                if channels and chId not in channels:
                    yield None
                    continue

                # TODO: Modulus correction, if still needed.
                scalar = timeScalars.get(chId, 1)
                blockStart *= scalar
                blockEnd *= scalar

                writeCurrent = True
                writePrev = False  # write previous block, if current one starts late

                if blockEnd < startTime:
                    # Entirely before extraction interval. Yield nothing.
                    writeCurrent = False
                elif blockStart <= startTime:
                    # Block overlaps start of interval. Yield block.
                    # Mark channel finished if block also includes end of interval.
                    writePrev = False
                    finished[chId] = blockEnd >= endTime
                elif blockEnd <= endTime:
                    # Block within interval. Write block (w/ prev. if needed).
                    writePrev = channelsWritten[chId] == 0
                else:
                    # Block overlaps end of interval. Yield block (w/ prev. if needed).
                    # Mark channel finished.
                    finished[chId] = True
                    writePrev = channelsWritten[chId] == 0

                if writePrev:
                    # Yield the previous block, which is before the extraction interval.
                    # This is to ensure that initial data is not left out.
                    prev = lastBlocks.get(chId, None)
                    if prev:
                        channelsWritten[chId] += 1
                        yield prev[0]

                lastBlocks[chId] = (el, blockStart, blockEnd)

                if writeCurrent:
                    channelsWritten[chId] += 1
                    yield el
                else:
                    yield None

            else:
                # FUTURE: Omit `<Sync>` elements outside the interval and
                #  'manually' create ones immediately before and after? Omit
                #  certain time-specific `<Attribute>` elements as well
                #  (if any)?
                yield el

    except (StopIteration, KeyboardInterrupt):
        pass


def readData(doc, source=None, startTime=None, endTime=None, channels=None,
             updater=None, total=None, bytesRead=0, samplesRead=0,
             parserTypes=None, **kwargs):
    """ Import the data from a file into a Dataset.
    
        :param doc: The Dataset document into which to import the data.
        :param source: An alternate Dataset to merge into the main one.
        :param startTime: The start of the extraction range, relative to the
            recording's start.
        :param endTime: The end of the extraction range, relative to the
            recording's end.
        :param channels: A list of channel IDs to import. If `None` (the
            default), all channels are imported.
        :param updater: A function (or function-like object) to notify as
            work is done. It should take four keyword arguments: `count` (the 
            current line number), `total` (the total number of samples), `error` 
            (an unexpected exception, if raised during the import), and `done` 
            (will be `True` when the export is complete). If the updater object 
            has a `cancelled` attribute that is `True`, the import will be 
            aborted. The default callback is `None` (nothing will be notified).
        :param total: The total number of bytes in the file(s) being imported.
            Defaults to the size of the current file, but can be used to
            display an overall progress when merging multiple recordings. For
            display purposes.
        :param bytesRead: The number of bytes already imported. Mainly for
            merging multiple recordings. For display purposes.
        :param samplesRead: The total number of samples imported. Mainly for
            merging multiple recordings.
        :param parserTypes: A collection of `parsers.ElementHandler` classes.
        :return: The total number of samples read.
    """
    kwargs.pop('sessionId', None)  # Unused; for Classic compatibility.

    # FUTURE: Remove `kwargs` and this conditional warning.
    if kwargs:
        warnings.warn(DeprecationWarning(
                'Some readData() updater-related arguments have been deprecated. '
                'Ignored arguments: {}'.format(', '.join(repr(k) for k in kwargs))))

    parserTypes = parserTypes or ELEMENT_PARSER_TYPES
    if doc._parsers is None:
        # Possibly redundant; is `doc._parsers` ever `None` at this point?
        doc._parsers = instantiateParsers(doc, parserTypes)
    
    elementParsers = doc._parsers

    elementCount = 0
    numSamples = 0  # Number of samples imported from this file

    # Progress display setup
    if updater and total is None:
        total = _getSize(doc.ebmldoc.stream) + bytesRead

    increment = 50  # Number of elements per updater. FUTURE: Base this on total size?
    timeOffset = 0

    # Actual importing ---------------------------------------------------------
    if source is None:
        source = doc

    try:
        if startTime is not None or endTime is not None or channels:
            iterator = filterTime(doc, startTime, endTime, channels=channels)
        else:
            iterator = iter(source.ebmldoc)

        for n, el in enumerate(iterator):
            # Progress display stuff -------------------------------------
            if updater:
                loadCancelled = getattr(updater, "cancelled", False)
                if loadCancelled:
                    doc.loadCancelled = loadCancelled
                    break

                thisOffset = el.offset + bytesRead
                if n % increment == 0:
                    updater(count=numSamples + samplesRead,
                            percent=thisOffset / total)

            if el is None:
                continue

            el_name = el.name

            if el_name not in elementParsers:
                # Unknown block type; probably okay to skip.
                logger.info("unknown block {!r} (ID 0x{:02x}) @{}".format(
                        el_name, el.id, el.offset))
                continue

            if source != doc and el_name == "TimeBaseUTC":
                timeOffset = (el.value - doc.lastSession.utcStartTime) * 1000000.0
                continue
                
            try:
                parser = elementParsers[el_name]

                # "Header" elements were loaded by `openFile()`; don't duplicate.
                if not parser.isHeader or el_name == "Attribute":
                    added = parser.parse(el, timeOffset=timeOffset)
                    if isinstance(added, int):
                        numSamples += added
                    
            except parsers.ParsingError as err:
                # TODO: Error messages?
                logger.error("Parsing error during import: %s" % err)
                continue

            elementCount += 1

    except IOError as e:
        if e.errno is None:
            # The EBML library raises an empty IOError if it hits EOF.
            # TODO: Verify that this still does anything (leftover from old EBML library)
            doc.fileDamaged = True
        elif updater:
            updater(error=e, done=True)
        
    except TypeError:
        # This can occur if there is a bad element in the data
        # (typically the last)
        doc.fileDamaged = True

    doc.fillCaches()
    doc.loading = False

    if updater:
        updater(done=True)

    return numSamples




"""
Calibration and Transform classes. These are callable objects that modify
data. Functionally, calibration is the same as a Transform, the difference
being in how they are used.

:author: dstokes
"""

__all__ = ['Transform', 'Univariate', 'Bivariate', 'CombinedPoly', 'PolyPoly',
           'AccelTransform']

import weakref
from collections import OrderedDict
import math
from time import sleep
import warnings

import logging

import numpy as np

logger = logging.getLogger('idelib')
logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s")

#===============================================================================
# 
#===============================================================================


class Transform(object):
    """ Base class for all data-manipulating objects (e.g. calibration
        polynomials). Instantiates as a function-like object representing any 
        processing that an event requires, including basic calibration at the 
        low level and adjustments for display at the high level.
    """
    modifiesTime = False
    modifiesValue = False
    units = None
    
    def __init__(self, *args, **kwargs):
        self.id = None
        self._str = "x"
        self._source = "lambda x: x"
        self._function = eval(self._source, {'math': math})
        self._variables = ("x",)
        self._lastSession = None
        self._timeOffset = 0
        self._watchers = weakref.WeakSet()
        
        # Custom attributes, e.g. Attribute elements in the EBML.
        self.attributes = kwargs.pop('attributes', None)
    
    
    def copy(self):
        """ Create a duplicate of this Transform.
        """
        t = self.__class__()
        for attr in ('id', '_str', '_source', '_function', '_lastSession', 
                     '_timeOffset'):
            setattr(t, attr, getattr(self, attr, None))
        return t
    
    
    def __str__(self):
        return self._str
    
    
    def __repr__(self):
        cname = self.__class__.__name__
        if self.id is None:
            return "<%s: (%s)>" % (cname, self._str)
        return "<%s (ID %d): (%s)>" % (cname, self.id, self._str)


    def __hash__(self):
        return hash(self._str)


    def __eq__(self, other):
        try:
            return self._str == other._str
        except AttributeError:
            return False
    
    
    def __ne__(self, other):
        return not self.__eq__(other)


    @property
    def function(self):
        """ The generated polynomial function itself. """
        return self._function


    @property
    def source(self):
        """ The optimized source code of the polynomial. """
        return self._source
    
    
    def __call__(self, timestamp, value, session=None, noBivariates=False):
        if session != self._lastSession:
            self._timeOffset = 0 if session.startTime is None else session.startTime
            self._session = session
        return timestamp + self._timeOffset, self._function(value)


    def isValid(self, session=None, noBivariates=False):
        """ Check the validity of the Transform.
        """
        # TODO: More base validity tests for all Transform subclasses?
        try:
            return (self.function is not None)
        except AttributeError:
            return False


    def inplace(self, values, y=None, timestamp=None, session=None, noBivariates=False, out=None):
        """ In-place Transform call for the Transform base class.
            All subclasses are required to implement this method, so it will
            raise an error if called.
        """
        raise NotImplementedError()


    @property
    def useMean(self):
        """ Property to determine if a transform will use the mean of the
            secondary channel.  The base class and all univariate polynomials
            do not have a secondary channel, so they will always return the
            default `True` and generate a warning.
        """
        warnings.warn(UserWarning('{} does not support useMean'.format(type(self))))
        return True


    @useMean.setter
    def useMean(self, value):
        warnings.warn(UserWarning('{} does not support useMean'.format(type(self))))


    def addWatcher(self, watcher):
        """ Adds `watcher` to the list (a `WeakSet`) of watchers.  The watchers
            are other polynomials, such as a `CombinedPoly`, which reference
            this polynomial.  This is used to propogate changes to all
            polynomials which reference this one.
        """
        self._watchers.add(watcher)

    
    #===========================================================================
    # 
    #===========================================================================

    @classmethod
    def null(cls, *args, **kwargs):
        return args[0]


class ComplexTransform(Transform):
    """ An mixin class for arbitrarily complex functions that don't follow a set form.

        Instances are function-like objects that take one argument: a sensor
        reading.
    """

    def inplace(self, values, y=None, timestamp=None, session=None, noBivariates=False, out=None):
        """ In-place transform for the `ComplexTransform` transform.  These functions
            can't be easily reduced and combined like polynomials, so this instead
            wraps the `function` method.
        """

        scalar = np.isscalar(values)

        if scalar:
            values = float(values)
        elif out is None:
            out = np.zeros_like(values, dtype=np.float64)

        if len(self._variables) == 1:

            if scalar:
                out = self.function(values)
            else:
                out[:] = self.function(values)
        else:

            session = self.dataset.lastSession if session is None else session
            sessionId = None if session is None else session.sessionId

            try:
                if self._eventlist is None or self._sessionId != sessionId:
                    channel = self.dataset.channels[self.channelId][self.subchannelId]
                    self._eventlist = channel.getSession(session.sessionId)
                    self._sessionId = session.sessionId

            except IndexError as err:
                # In multithreaded environments, there's a rare race condition
                # in which the main channel can be accessed before the calibration
                # channel has loaded. This should fix it.
                logger.warning("%s occurred in Bivariate polynomial %r" %
                               err.__class__.__name__, self.id)
                return None

            if noBivariates:
                y = 0
            elif self.useMean:
                y = self._eventlist.getMean()
            elif y is None and timestamp is None:
                y = self._eventlist.getMean()
            elif y is None and timestamp is not None:
                y = np.fromiter((self._eventlist[self._eventlist.getEventIndexNear(t)][0] for t in
                                 timestamp), dtype=np.float64)

            if scalar:
                out = self.function(values, y)
            else:
                out[:] = self.function(values, y)

        return out



#===============================================================================
# Simple Transforms. These could be represented by Univariate polynomials,
# but are hard-coded for efficiency's sake.
#===============================================================================


class AccelTransform(Transform):
    """ A simple transform to convert accelerometer values (parsed as
        int16) to floats in the range -100 to 100 G.
         
        This assumes that the data was parsed by `AccelerometerParser`, which
        puts the raw values in the range -32768 to 32767.
    """
    modifiesValue = True
     
    def __init__(self, amin=-100, amax=100, calId=0, dataset=None):
        self.id = calId
        self.range = (amin, amax)
        self._str = "(x / 32767.0) * %.3f" % amax
        self._source = "lambda x: x * %f" % (amax / 32767.0)
        self._function = eval(self._source, {'math': math})
        self._lastSession = None
        self._timeOffset = 0
        
        self.references = (0,)
        self.coefficients = ((amax / 32767.0), amax)

        self.dataset = None
        self._watchers = weakref.WeakSet()


    def copy(self):
        """ Create a duplicate of this Transform.
        """
        t = self.__class__(self.range[0], self.range[1], self.id, self.dataset)
        return t


#===============================================================================
# Polynomial Generators
#===============================================================================


class Univariate(Transform):
    """ A simple calibration polynomial in the general form:
        `y=(coeffs[0]*(x**n))+(coeffs[1]*(x**n-1))+...+(coeffs[n]+(x**0))`.
        
        Instances are function-like objects that take one argument: a sensor
        reading.

        TODO: in-place changes force this to be in the form of either
              `f(x) = coeffs[0]*x + coeffs[1]`
              or
              `f(x) = coeffs[0]`
              This behavior will need to be updated later.
    """
    modifiesValue = True
    
    @classmethod
    def _floatOrInt(cls, v):
        """ Helper method to convert floats with no decimal component to ints. """
#         iv = int(v)
#         return iv if iv == v else v
        # Not a good optimization. Only convert 0 to integer.
        return 0 if v == 0 else float(v)
    
    
    @classmethod
    def _stremove(cls, s, old):
        """ Helper method to remove a set of substrings from a string. """
        result = str(s)
        for o in old:
            result = result.replace(o, '')
        return result
    
    
    @classmethod
    def _streplace(cls, s, *args):
        """ Helper method to replace multiple substrings. """
        for old, new in args:
            s = s.replace(old, new)
        return s
    
    
    @classmethod
    def _fixSums(cls, s):
        """ Helper method to replace consecutive addition/subtraction combos. """
        result = str(s)
        for old, new in (("--", "+"), ("-+", "-"), ("+-", "-"), ("++", "+")):
            result = result.replace(old, new)
        return result.lstrip('+')
    
    
    # XXX: kwargs added to __init__() as work-around for bad SSS templates! 
    def __init__(self, coeffs, calId=None, dataset=None, reference=0, 
                 varName="x", attributes=None, **kwargs):
        """ Construct a simple polynomial function from a set of coefficients.
            
            :param coeffs: A list of coefficients
            :keyword calId: The polynomial's calibration ID (if any).
            :keyword dataset: The parent `dataset.Dataset`.
            :keyword reference: A reference value to be subtracted from the
                variable.
            :keyword varName: The name of the variable to be used in the
                string version of the polynomial. For display purposes.
            :keyword attributes: A dictionary of generic attributes, e.g.
                ones parsed from `Attribute` EBML elements, or `None`.
        """
        if len(coeffs) > 2 or len(coeffs) == 0:
            raise ValueError('Coefficients of length {} are not supported.  '
                             'Coefficients must be length 1 (constant) or 2 '
                             '(linear).'.format(len(coeffs)))
        coeffs = tuple(float(x) for x in coeffs)
        self.id = calId
        self.dataset = dataset
        self._coeffs = tuple(coeffs)
        self._fastCoeffs = tuple(coeffs)
        self._variables = (varName,)
        self._references = (reference,)
        self._session = None
        self._lastSession = None
        self._timeOffset = 0
        self._watchers = weakref.WeakSet()
        
        self.attributes = attributes
        
        self._build()


    def __hash__(self):
        return hash((self._coeffs, self._references, self.id))
        

    def __eq__(self, other):
        try:
            return self._coeffs == other._coeffs and \
                   self._references == other._references and \
                   self.id == other.id
        except AttributeError:
            return False


    def copy(self):
        """ Create a duplicate of this Transform.
        """
        return self.__class__(self._coeffs, calId=self.id, dataset=self.dataset,
                              reference=self._references[0], 
                              varName=self._variables[0])

    
    def _build(self):
        """ Internal method that (re-)constructs the polynomial function.
        """
        varName = str(self._variables[0])
        srcVarName = "x"
        reference = self._references[0]
        coeffs = self._coeffs
        
        if reference != 0:
            varName = "(%s-%s)" % (varName, reference)
            srcVarName = "(%s-%s)" % (srcVarName, reference)

        if len(coeffs) == 1:
            self._fastCoeffs = tuple(self._coeffs)
            self._str = str(self._fastCoeffs[0])
        elif len(coeffs) == 2:
            # a0*(x - ref) + a1 = a0*x + (a1 - a0*ref)
            self._fastCoeffs = (self._coeffs[0], self._coeffs[1] - self._coeffs[0]*reference)
        else:
            raise ValueError('Invalid list of coefficients.')


        # f is used to build the lambda
        # strF is used to build the string version
        coeffs = list(reversed(coeffs))
        f = [coeffs[0]] if coeffs[0] != 0 else []
        strF = f[:]
        coeffs = [self._floatOrInt(c) for c in coeffs]
        for p, v in enumerate(coeffs[1:], 1):
            # optimization: x*0 == 0
            if v == 0:
                continue
    
            # optimization: pow() is more expensive than lots of multiplication
            x = "*".join([srcVarName]*p)
            strX = "pow(%s,%s)" % (varName, p) if p > 1 else varName
    
            # optimization: remove multiplication by 1
            q = "%s*" % v if v != 1 else ""
    
            f.append("(%s%s)" % (q, x))
            strF.append("(%s%s)" % (q, strX))

        if tuple(coeffs) == (0,):
            f = [0]
            strF = [0]
    
        self._str = "+".join(map(str, reversed(strF)))
        self._str = self._fixSums(self._str)
        
        self._source = "lambda x: %s" % ("+".join(map(str, reversed(f))))
        self._source = self._fixSums(self._source)
        self._function = eval(self._source, {'math': math})


    def inplace(self, values, y=None, timestamp=None, session=None, noBivariates=False, out=None):
        """ In-place transform for the `Univariate` transform.  It reduces the
            number of array allocations/operations compared to the normal
            `__call__` method.  The user can supply an `out` argument to save
            the results to that array, or leave `out` equal to `None`.  If `out`
            is `None`, then this method will allocate an array in the shape of
            `values`.

            The method uses the following algebra, calculated in the `_build`
            method, to reduce the equation to fewer operations.  For brevity:
                `x` = `values`
                `an` = `coeffs[n]`
                `bn` = `_fastCoeffs[n]`

            `f(x) = a0*(x - ref) + a1`
             =>
            `f(x) = x*(a0) + (a1 - a0*ref)`
             =>
            `f(x) = b0*x + b1`

        """

        scalar = np.isscalar(values)

        if scalar:
            values = float(values)
        elif out is None:
            out = np.zeros_like(values, dtype=np.float64)

        if len(self._fastCoeffs) == 1:
            if scalar:
                out = self._fastCoeffs[0]
            else:
                out[:] = self._fastCoeffs[0]
        elif len(self._fastCoeffs) == 2:
            if scalar:
                out = values
            else:
                out[:] = values
            out *= self._fastCoeffs[0]
            out += self._fastCoeffs[1]
        else:
            raise

        return out
    
    
    @property
    def coefficients(self):
        """ The polynomial's coefficients. """
        return self._coeffs
    
    
    @coefficients.setter
    def coefficients(self, val):
        self._coeffs = val
        self._build()
        
        
    @property
    def variables(self):
        """ The name(s) of the variable(s) used in the polynomial. """
        return self._variables
    
    
    @variables.setter
    def variables(self, val):
        self._variables = val
        self._build()
    

    @property
    def references(self):
        """ The constant offset(s). """
        return self._references
    
    
    @references.setter
    def references(self, val):
        self._references = val
        self._build()
    
    
    def asDict(self):
        """ Dump the polynomial as a dictionary. Intended for use when
            generating EBML.
        """
        return OrderedDict((('CalID', self.id),
                            ('CalReferenceValue', self._references[0]),
                            ('PolynomialCoef', self._coeffs)))


#===============================================================================
# 
#===============================================================================

class Bivariate(Univariate):
    """ A two-variable polynomial in the general form
        `v=(A*x*y)+(B*x)+(C*y)+D`.
        
        Instances are function-like objects that take one argument: a sensor
        reading time and value.
    """
    
    def __init__(self, coeffs, calId=None, dataset=None, reference=0, 
                 reference2=0, channelId=None, subchannelId=None, varNames="xy",
                 attributes=None):
        """ Construct the two-variable polynomial.
            
            :param coeffs: A list of coefficients. Must contain 4!
            :keyword reference: A reference value to be subtracted from the
                'x' variable.
            :keyword reference2: A reference value to be subtracted from the
                'y' variable.
            :keyword varNames: The names of the variables to be used in the
                string version of the polynomial. For display purposes; they
                can be any arbitrary (but hopefully meaningful) strings.
            :keyword calId: The polynomial's calibration ID (if any).
            :keyword dataset: The parent `dataset.Dataset`.
            :keyword attributes: A dictionary of generic attributes, e.g.
                ones parsed from `Attribute` EBML elements, or `None`.
        """
        if len(coeffs) not in [1, 4]:
            raise ValueError('Coefficients of length {} are not supported.  '
                             'Coefficients must be length 1 (constant) or 4 '
                             '(linear).'.format(len(coeffs)))
        coeffs = tuple(float(x) for x in coeffs)
        self.dataset = dataset
        self._eventlist = None
        self._sessionId = None
        self.channelId, self.subchannelId = channelId, subchannelId
        if channelId is None or subchannelId is None:
            raise ValueError("Bivariate polynomial requires channel and "
                    "subchannel IDs; got %r, %d" % (channelId, subchannelId))
        
        self.id = calId
        
        self._references = (float(reference), float(reference2))
        self._coeffs = tuple(map(float, coeffs))
        self._fastCoeffs = tuple(self._coeffs)
        self._variables = tuple(map(str, varNames))
        
        self._session = None
        self._timeOffset = 0
        
        self.attributes = attributes

        self._useMean = True
        self._watchers = weakref.WeakSet()
        
        self._build()


    def __hash__(self):
        return hash(self._str)
        
        
    def __eq__(self, other):
        try:
            return self._str == other._str
        except AttributeError:
            return False
    
    
    def copy(self):
        """ Create a duplicate of this Transform.
        """
        # This could be optimized by circumventing the polynomial rebuild,
        # and instead using the already-generated source.
        return self.__class__(self._coeffs, dataset=self.dataset, 
               channelId=self.channelId, subchannelId=self.subchannelId, 
               reference=self._references[0], reference2=self._references[1], 
               varNames=self._variables, calId=self.id)

    
    @classmethod
    def _reduce(cls, src):
        """ Simple reduction for polynomial expressions, removing redundant
            parts (adding/subtracting 0, multiplying by 0 or 1, etc.).
        """
        old = None
        while old != src:
            old = src
            src = cls._streplace(src, 
                ('(0+', '('), ('(0.0+', '('), ('(0-', '(-'), ('(0.0-', '(-'),
                ('(0*x', '(0'), ('(0.0*x', '(0'), 
                ('(0*y', '(0'), ('(0.0*y', '(0'),
                ('(0*x*y)+', ''), ('(0*x)+', ''), ('(0*y)+', ''),
                ('*)', ')'), ('+)', ')'), 
                ('(0)', ''), ('(0.0)', ''), ('()', ''), 
                ('(1)', '1.0'), ('(1.0)', '1.0'),
                ("(1*", "("), ("(1.0*", "("),  # ("(x)", "x"), ("(y)", "y"),
                ("--", "+"), ("-+", "-"), ("+-", "-"), ("++", "+")
            )
            if src.endswith('+0') or src.endswith('+0.0'):
                src = src.rstrip('+0.')
        return src
    
    
    def _build(self):
        """ Internal method that (re-)constructs the polynomial function.
        """
        coeffs = self._coeffs
        reference, reference2 = self._references
        varNames = self._variables

        if len(coeffs) not in [1, 4]:
            raise ValueError('Coefficients of length {} are not supported.  '
                             'Coefficients must be length 1 (constant) or 4 '
                             '(linear).'.format(len(coeffs)))

        if len(coeffs) == 1:
            self._fastCoeffs = tuple(self._coeffs)
        if len(coeffs) == 4:
            # a0*(x - refx)*(y - refy) + a1*(x - refx) + a2*(y - refy) + a3 =
            # = a0*x*y + x*(a1 - a0*refy) + y*(a2 - a0*refx) +
            #   + (a0*refx*refy + a3 - a1*refx - a2*refy)
            self._fastCoeffs = (
                self._coeffs[0],
                self._coeffs[1] - self._coeffs[0]*reference2,
                self._coeffs[2] - self._coeffs[0]*reference,
                self._coeffs[3] + self._coeffs[0]*reference*reference2 -
                self._coeffs[1]*reference - self._coeffs[2]*reference2,
                )

        # Construct the polynomial expression string
        strF = "(%s*x*y)+(%s*x)+(%s*y)+%s"
        src = strF % tuple(map(self._floatOrInt, coeffs))

        # Build the display (i.e. unoptimized) version of the polynomial         
        self._str = strF % self._coeffs
        varNames = list(varNames)
        if reference != 0:
            varNames[0] = "(x-%s)" % reference
        if reference2 != 0:
            varNames[1] = "(y-%s)" % reference2
        self._str = self._fixSums(self._str)
        
        # Replace standard variable names with custom ones. Change one to 
        # a stand-in first, to prevent problems if one name contains the other.
        self._str = self._streplace(self._str, ("x", "\x00"), 
                                    ("y", varNames[1]), ("\x00", varNames[0]))

        # Optimizations: Build a simplified expression for function.
        # 1. Remove multiples of 0 and 1, addition of 0 constants.
        src = self._reduce(src)    

        # 2. If there's a reference value, replace the variable with (v-ref)
        references = [self._floatOrInt(x) for x in (reference, reference2)]
        for i, v in enumerate("xy"):
            if references[i] != 0:
                src = src.replace(v, "(%s-%s)" % (v, references[i]))  
        
        # 3. Do the reduction again, now that the offsets are in.
        src = self._reduce(src)    
        if not src:
            logger.warning("Bad polynomial coefficients {}".format(self._coeffs))
            src = "x"
        
        self._source = 'lambda x,y: %s' % self._fixSums(src)
        self._function = eval(self._source, {'math': math})
        
        # Optimization: it is possible that the polynomial could exclude Y
        # completely. If that's the case, use a dummy value to speed things up.
        self._noY = (0, 1) if 'y' not in src else False


    def inplace(self, values, y=None, timestamp=None, session=None, noBivariates=False, out=None):
        """ In-place transform for the `Bivariate` transform.  It reduces the
            number of array allocations/operations compared to the normal
            `__call__` method.  The user can supply an `out` argument to save
            the results to that array, or leave `out` equal to `None`.  If `out`
            is `None`, then this method will allocate an array in the shape of
            `values`.

            When calculating the `y` value, several things need to be taken into
            account in the following priority:
            1. If `noBivariates` is `True`, then `y` must be equal to `0`.
            2. If `self.useMean` is `True`, then `y` is equal to the mean of the
               full length of the secondary channel.
            3. If `y` is provided, then use the provided value.
            4. If `timestamp` is provided, then get the values for the given `timestamp`.
            5. Otherwise, default to the mean of the secondary channel.

            Depending on if `y` is scalar, one of two methods will be used, detailed below.

            The method uses the following algebra, calculated in the `_build`
            method, to reduce the equation to fewer operations.  For brevity:
                `x` = `values`
                `xref`, `yref` = `references`
                `an` = `coeffs[n]`
                `bn` = `_fastCoeffs[n]`

            `y` is not scalar:

              `f(x) = a0*(x - xref)*(y - yref) + a1*(x - xref) + a2*(y - yref) + a3`
               =>
              `f(x) = x*y*(a0) + x*(a1 - a0*yref) + y*(a2 - a0*xref) + (a3 + a0*xref*yref - a1*xref - a2*yref)`
               =>
              `f(x) = b0*x*y + b1*x + b2*y + b3`

            `y` is scalar:

              `f(x) = b0*x*y + b1*x + b2*y + b3`
               =>
              `f(x) = x*(b0*y + b1) + (b2*y + b3)`
        """

        scalar = np.isscalar(values)

        if scalar:
            values = float(values)
        elif out is None:
            out = np.zeros_like(values, dtype=np.float64)

        session = self.dataset.lastSession if session is None else session
        sessionId = None if session is None else session.sessionId

        try:
            if self._eventlist is None or self._sessionId != sessionId:
                channel = self.dataset.channels[self.channelId][self.subchannelId]
                self._eventlist = channel.getSession(session.sessionId)
                self._sessionId = session.sessionId

        except IndexError as err:
            # In multithreaded environments, there's a rare race condition
            # in which the main channel can be accessed before the calibration
            # channel has loaded. This should fix it.
            logger.warning("%s occurred in Bivariate polynomial %r" %
                           err.__class__.__name__, self.id)
            return None

        if noBivariates:
            y = 0
        elif self.useMean:
            y = self._eventlist.getMean()
        elif y is None and timestamp is None:
            y = self._eventlist.getMean()
        elif y is None and timestamp is not None:
            y = np.fromiter((self._eventlist[self._eventlist.getEventIndexNear(t)][0] for t in timestamp), dtype=np.float64)

        if len(self._fastCoeffs) == 1:
            if scalar:
                out = self._fastCoeffs[0]
            else:
                out[:] = self._fastCoeffs[0]
        elif len(self._fastCoeffs) == 4:
            if np.isscalar(y):
                # a0*x*y + a1*x + a2*y + a3 =
                # x*(a0*y + a1) + (a2*y + a3)
                if scalar:
                    out = values
                else:
                    out[:] = values
                out *= self._fastCoeffs[0]*y + self._fastCoeffs[1]
                out += self._fastCoeffs[2]*y + self._fastCoeffs[3]
            else:
                out += values*y*self._fastCoeffs[0]
                out += values*self._fastCoeffs[1]
                out += y*self._fastCoeffs[2]
                out += self._fastCoeffs[3]

        else:
            raise

        return out


    def __call__(self, timestamp, value, session=None, noBivariates=False):
        """ Apply the polynomial to an event. 
        
            :param timestamp: The time of the event to process.
            :param value: The value of the event to process.
            :keyword session: The session containing the event.
            :keyword noBivariates: If `True`, the reference channel will not
                be used.
        """
        session = self.dataset.lastSession if session is None else session
        sessionId = None if session is None else session.sessionId
        
        try:
            if self._eventlist is None or self._sessionId != sessionId:
                channel = self.dataset.channels[self.channelId][self.subchannelId]
                self._eventlist = channel.getSession(session.sessionId)
                self._sessionId = session.sessionId
            if len(self._eventlist) == 0:
                return timestamp, value
            
            # Optimization: don't check the other channel if Y is unused
            if noBivariates:
                y = (0, 1)
            else:
                y = self._noY or self._eventlist.getMean()
            return timestamp, self._function(value, y)
        
        except (IndexError, ZeroDivisionError) as err:
            # In multithreaded environments, there's a rare race condition
            # in which the main channel can be accessed before the calibration
            # channel has loaded. This should fix it.
            logger.warning("%s occurred in Bivariate polynomial %r" %
                           err.__class__.__name__, self.id)
            return None

        except (IndexError, ZeroDivisionError) as err:
            # In multithreaded environments, there's a rare race condition
            # in which the main channel can be accessed before the calibration
            # channel has loaded. This should fix it.
            logger.warning("%s occurred in Bivariate polynomial %r" %
                           err.__class__.__name__, self.id)
            return None


    def asDict(self):
        """ Dump the polynomial as a dictionary. Intended for use when
            generating EBML.
        """
        cal = super(Bivariate, self).asDict()
        cal['BivariateCalReferenceValue'] = self._references[1]
        cal['BivariateChannelIDRef'] = self.channelId
        cal['BivariateSubChannelIDRef'] = self.subchannelId
        return cal


    def isValid(self, session=None, noBivariates=False, _retries=3):
        """ Check the validity of the Transform.
        
            :keyword session: The session to check (could be valid in one and
                invalid in another, e.g. one session has no temperature data).
            :keyword noBivariates: If `True`, the reference channel will not
                be used.
        """
        valid = super(Bivariate, self).isValid(session, noBivariates)
        if noBivariates or not valid:
            return valid
        
        session = self.dataset.lastSession if session is None else session
        sessionId = None if session is None else session.sessionId
        
        try:
            if self._eventlist is None or self._sessionId != sessionId:
                channel = self.dataset.channels[self.channelId][self.subchannelId]
                self._eventlist = channel.getSession(session.sessionId)
                self._sessionId = session.sessionId
            if len(self._eventlist) == 0:
                return False
            
        except:
            # HACK: In multithreaded environments, there's a rare race 
            # condition in which the main channel can be accessed before the
            # calibration channel has loaded. Retry isValid() a few times.
            if _retries == 0:
                return False
            
            return self.isValid(session, noBivariates, _retries-1)


    @property
    def useMean(self):
        return self._useMean


    @useMean.setter
    def useMean(self, value):
        self._useMean = value
        for w in self._watchers:
            w.useMean = self.useMean


#===============================================================================
# 
#===============================================================================

class CombinedPoly(Bivariate):
    """ Calibration transform that combines multiple polynomials into a single
        function. Used for combining Channel and Subchannel transforms to make
        them more efficient.
    """
    
    def copy(self):
        """ Create a duplicate of this Transform.
        """
        return self.__class__(self.poly, subchannel=self._subchannel, 
                              calId=self.id, dataset=self.dataset, 
                              **self.kwargs)
    
    def __init__(self, poly, subchannel=None, calId=None, dataset=None, 
                 **kwargs):
        self.id = calId
        self.poly = poly
        self.subpolys = kwargs
        self._fastCoeffs = None
        
        if subchannel is not None:
            self.poly = self.poly[subchannel]

        self.kwargs = kwargs
        p = list(kwargs.values())[0]
        for attr in ('dataset', '_eventlist', '_sessionId', 'channelId',
                     'subchannelId', '_references', '_coeffs', '_variables'):
            setattr(self, attr, getattr(poly, attr, getattr(p, attr, None)))
                    
        self.dataset = self.dataset or dataset
        self._subchannel = subchannel
        self._watchers = weakref.WeakSet()
        self._build()

        self._useMean = True
    
    
    def _build(self):
        if self.poly is None:
            self.poly = Univariate((1, 0))
            if len(self.subpolys) == 1:
                p = list(self.subpolys.values())[0]
                if p is not None:
                    for attr in ('_str', '_source', '_function', '_noY'):
                        setattr(self, attr, getattr(p, attr, None))
        phead, _, src = self.poly.source.partition(": ")

        for k, v in self.subpolys.items():
            if v is None:
                ssrc = "lambda x: x"
            else:
                ssrc = v.source
            start, _, end = ssrc.rpartition(": ")
            s = "(%s)" % (end or start)
            src = self._reduce(src.replace(k, s))

        if self._subchannel is not None:
            src = src.replace('x', 'x[%d]' % self._subchannel)
        
        # Merge in all function globals, in case components use additional
        # libraries (e.g. math). 
        evalGlobals = {'math': math}
        if self.poly is not None: 
            evalGlobals.update(self.poly._function.__globals__)
        for p in self.subpolys.values():
            if p is not None:
                evalGlobals.update(p._function.__globals__)

        if isinstance(self.poly, Bivariate):
            # a0*(b0*x + b1)*(c0*y + c1) + a1*(b0*x + b1) + a2*(c0*y + c1) + a3 =
            # x*y*(a0*b0*c0) + x*(a0*b0*c1 + a1*b0) + y*(a0*b1*c0 + a2*c0) + (a0*b1*c1 + a1*b1 + a2*c1 + a3)
            a = self.poly._fastCoeffs
            if self.poly.variables[0] in self.subpolys:
                if self.subpolys[self.variables[0]] is not None and \
                        not isinstance(self.subpolys[self.variables[0]], ComplexTransform):
                    b = self.subpolys[self.variables[0]]._fastCoeffs
                else:
                    b = (1, 0)
            else:
                b = (1, 0)
            if self.poly.variables[1] in self.subpolys:
                if self.subpolys[self.variables[1]] is not None and \
                        not isinstance(self.subpolys[self.variables[1]], ComplexTransform):
                    c = self.subpolys[self.variables[1]]._fastCoeffs
                else:
                    c = (1, 0)
            else:
                c = (1, 0)
            self._fastCoeffs = (
                a[0]*b[0]*c[0],
                a[0]*b[0]*c[1] + a[1]*b[0],
                a[0]*b[1]*c[0] + a[2]*c[0],
                a[0]*b[1]*c[1] + a[1]*b[1] + a[2]*c[1] + a[3],
                )
        elif isinstance(self.poly, Univariate):
            # a0*(b0*x + b1) + a1 = a0*b0*x + (a1 + a0*b1)
            a = self.poly._fastCoeffs
            b = self.subpolys[self.poly.variables[0]]
            if b is None:
                b = (1., 0.)
            else:
                b = b._fastCoeffs
            self._fastCoeffs = (a[0]*b[0], a[1] + a[0]*b[1])
        
        self._str = src
        self._source = "%s: %s" % (phead, src)
        self._function = eval(self._source, evalGlobals)
        self._noY = (0, 1) if 'y' not in src else False

        for x in [self.poly] + list(self.subpolys.values()):
            if x is not None:
                x.addWatcher(self)

    def inplace(self, values, y=None, timestamp=None, session=None, noBivariates=False, out=None):
        """ In-place transform for the `CombinedPoly` transform.  It reduces the
            number of array allocations/operations compared to the normal
            `__call__` method.  The user can supply an `out` argument to save
            the results to that array, or leave `out` equal to `None`.  If `out`
            is `None`, then this method will allocate an array in the shape of
            `values`.

            If any component polynomials are a `ComplexTransform`, those will be
            calculated first.

            If the main polynomial is a `Complex Transform`, then all values
            will be calculated separately.

            If the component polynomials of this `CombinedPoly` are all
            `Univariate`, then the following equation will be used.  Otherwise,
            the rest of the documentation will detail how a `CombinedPoly` based
            on a `Bivariate` behaves.

            The method uses the following algebra, calculated in the `_build`
            method, to reduce the equation to fewer operations.  For brevity:
                `x` = `values`
                `xref`, `yref` = `references`
                `an` = `poly.coeffs[n]`
                `bn` = `subpoly['x']._fastCoeffs[n]`
                `cn` = `_fastCoeffs[n]`

            `f(x) = a0*(b0*x + b1 - xref) + a1`
             =>
            `f(x) = x*(a0*b0) + (a0*(b1 - xref) + a1)`
             =>
            `f(x) = c0*x + c1`

            When calculating the `y` value, several things need to be taken into
            account in the following priority:
            1. If `noBivariates` is `True`, then `y` must be equal to `0`.
            2. If `self.useMean` is `True`, then `y` is equal to the mean of the
               full length of the secondary channel.
            3. If `y` is provided, then use the provided value.
            4. If `timestamp` is provided, then get the values for the given `timestamp`.
            5. Otherwise, default to the mean of the secondary channel.

            Depending on if `y` is scalar, one of two methods will be used, detailed below.

            The method uses the following algebra, calculated in the `_build`
            method, to reduce the equation to fewer operations.  For brevity:
                `x` = `values`
                `xref`, `yref` = `references`
                `an` = `poly.coeffs[n]`
                `bn` = `subpoly['x']._fastCoeffs[n]`
                `cn` = `subpoly['y']._fastCoeffs[n]`
                `dn` = `_fastCoeffs[n]`

            `y` is not scalar:

              `f(x) = a0*(b0*x + b1 - xref)*(c0*y + c1 - yref) + a1*(b0*x + b1 - xref) +
                    + a2*(c0*y + c1 - yref) + a3`
               =>
              `f(x) = x*y*(a0*b0*c0) + x*(a0*b0*(c1 - yref) + a1*b0) +
                    + y*(a0*(b1 - xref)*c0 + a2*c0) +
                    + (a0*(b1 - xref)*(c1 - yref) + a1*(b1 - xref) + a2*(c1 - yref) + a3)`
               =>
              `f(x) = d0*x*y + d1*x + d2*y + d3`

            `y` is scalar:

              `f(x) = d0*x*y + d1*x + d2*y + d3`
               =>
              `f(x) = x*(d0*y + d1) + (d2*y + d3)`
        """

        scalar = np.isscalar(values)

        # Catches the case where the first subpoly is a `ComplexTransform`
        if self.variables and \
                self.variables[0] in self.subpolys and \
                isinstance(self.subpolys[self.variables[0]], ComplexTransform):
            values = self.subpolys[self.variables[0]].inplace(
                    values,
                    timestamp=timestamp,
                    session=session,
                    noBivariates=noBivariates,
                    )

        if scalar:
            values = float(values)
        elif out is None:
            out = np.zeros_like(values, dtype=np.float64)

        if self.variables is None:
            if scalar:
                return values
            out[:] = values
            return out

        try:
            if len(self.variables) == 1:
                if isinstance(self.poly, ComplexTransform):

                    if self.variables[0] in self.subpolys and not isinstance(self.subpolys[self.variables[0]], ComplexTransform):
                        values = self.subpolys[self.variables[0]].inplace(
                                values,
                                timestamp=timestamp,
                                session=session,
                                noBivariates=noBivariates,
                                )

                    if scalar:
                        out = self.poly.function(values)
                    else:
                        out[:] = self.poly.function(values)
                else:
                    if scalar:
                        out = values
                    else:
                        out[:] = values

                    out *= self._fastCoeffs[0]
                    out += self._fastCoeffs[1]
            else:

                session = self.dataset.lastSession if session is None else session
                sessionId = None if session is None else session.sessionId

                if self._eventlist is None or self._sessionId != sessionId:
                    channel = self.dataset.channels[self.channelId][self.subchannelId]
                    self._eventlist = channel.getSession(session.sessionId)
                    self._sessionId = session.sessionId

                if noBivariates:
                    y = 0
                elif self.useMean:
                    y = self._eventlist.getMean()
                elif y is None and timestamp is None:
                    y = self._eventlist.getMean()
                elif y is None and timestamp is not None:
                    y = np.fromiter(
                            (self._eventlist[self._eventlist.getEventIndexNear(t)][0] for t in
                             timestamp), dtype=np.float64)

                # Catches the case where the secondary subpoly is a `ComplexTransform`
                if self.variables[1] in self.subpolys and isinstance(self.subpolys[self.variables[1]], ComplexTransform):
                    y = self.subpolys[self.variables[1]].inplace(
                            y,
                            timestamp=timestamp,
                            session=session,
                            noBivariates=noBivariates,
                            )

                # in this instance, basically give up on efficiency
                if isinstance(self.poly, ComplexTransform):

                    if self.variables[0] in self.subpolys and not isinstance(self.subpolys[self.variables[0]], ComplexTransform):
                        values = self.subpolys[self.variables[0]].inplace(
                                values,
                                timestamp=timestamp,
                                session=session,
                                noBivariates=noBivariates,
                                )

                    if self.variables[1] in self.subpolys and not isinstance(self.subpolys[self.variables[1]], ComplexTransform):
                        y = self.subpolys[self.variables[1]].inplace(
                                y,
                                timestamp=timestamp,
                                session=session,
                                noBivariates=noBivariates,
                                )

                    if scalar:
                        out = self.poly.inplace(
                                values,
                                y=y,
                                timestamp=timestamp,
                                session=session,
                                noBivariates=noBivariates,
                                )
                    else:

                        out[:] = self.poly.inplace(
                                values,
                                y=y,
                                timestamp=timestamp,
                                session=session,
                                noBivariates=noBivariates,
                                )

                else:

                    if np.isscalar(y):
                        # a2*x*y + b2*x + c2*y + d2 =
                        # x*(a2*y + b) + (c2*y + d2) =
                        # x*a3 + b3
                        if scalar:
                            out = values
                        else:
                            out[:] = values
                        a3 = self._fastCoeffs[0]*y + self._fastCoeffs[1]
                        b3 = self._fastCoeffs[2]*y + self._fastCoeffs[3]
                        out *= a3
                        out += b3
                    else:
                        if np.scalar:
                            out = values*y*self._fastCoeffs[0]
                        else:
                            out[:] = values*y*self._fastCoeffs[0]
                        out += values*self._fastCoeffs[1]
                        out += y*self._fastCoeffs[2]
                        out += self._fastCoeffs[3]

            return out

        except (IndexError, ZeroDivisionError) as err:
            # In multithreaded environments, there's a rare race condition
            # in which the main channel can be accessed before the calibration
            # channel has loaded. This should fix it.
            logger.warning("%s occurred in Bivariate polynomial %r" %
                           err.__class__.__name__, self.id)
            return None
        
        
    def asDict(self):
        """ Dump the polynomial as a dictionary. Intended for use when
            generating EBML.
        """
        raise TypeError("Can't generate dictionary for %s" %
                        self.__class__.__name__)

    
    def isValid(self, session=None, noBivariates=False):
        """ Check the validity of the Transform.
        """
        if not Transform.isValid(self, session, noBivariates):
            return False
        return all(p.isValid(session, noBivariates) for p in self.subpolys.values())
        

#------------------------------------------------------------------------------ 

class PolyPoly(CombinedPoly):
    """ Calibration transform that combines multiple subchannel polynomials 
        into one single function.
    """
    
    def copy(self):
        """ Create a duplicate of this Transform.
        """
        return self.__class__(self.polys, calId=self.id, dataset=self.dataset)
    
    
    def __init__(self, polys, calId=None, dataset=None):
        self.id = calId
        self.polys = polys
        poly = polys[0]
        for attr in ('dataset', '_eventlist', '_sessionId', 'channelId', 'subchannelId'):
            setattr(self, attr, getattr(poly, attr, None))
        
        self.dataset = self.dataset or dataset
        self._eventlist = None
        self._watchers = weakref.WeakSet()
        self._build()

        self._useMean = True


    def _build(self):
        params = []
        body = []
        for n, p in enumerate(self.polys):
            params.append('x%d' % n)
            if p is None:
                body.append('x%d' % n)
            else:
                start, _, end = p.source.rpartition(':')
                body.append((end or start).replace('x', 'x%d' % n))
        
        # ends with a comma to ensure a tuple is returned
        src = "(%s,)" % (', '.join(body))
        self._str = src

        if 'y' in src:
            params.insert(0, 'y')
            self._noY = False
        else:
            self._noY = (0, 1)
        
        # Merge in all function globals, in case components use additional
        # libraries (e.g. math). 
        evalGlobals = {'math': math}
        for p in self.polys:
            if p is not None:
                evalGlobals.update(p._function.__globals__)
                p.addWatcher(self)
        
        self._source = "lambda %s: %s" % (','.join(params), src)
        self._function = eval(self._source, evalGlobals)
        self._variables = params


    def __call__(self, timestamp, values, session=None, noBivariates=False):
        """ Apply the polynomial to an event. 
        
            :param timestamp: The time of the event to process.
            :param values: The values of the event to process.
            :keyword session: The session containing the event.
        """
        try:
            # Optimization: don't check the other channel if Y is unused
            if self._noY is False:
                if noBivariates:
                    return timestamp, self._function(0, *values)
                    
                session = self.dataset.lastSession if session is None else session
                sessionId = None if session is None else session.sessionId
                
                if self._eventlist is None or self._sessionId != sessionId:
                    channel = self.dataset.channels[self.channelId][self.subchannelId]
                    self._eventlist = channel.getSession(session.sessionId)
                    self._sessionId = session.sessionId
                
                # XXX: Hack! EventList length can be 0 if a thread is running.
                # This almost immediately gets fixed. Find real cause.
                try:    
                    y = self._eventlist.getMean()
                except IndexError:
                    sleep(0.001)
                    if len(self._eventlist) == 0:
                        return None
                    y = self._eventlist.getMean()
                    
                return timestamp, self._function(y, *values)
            
            else:
                return timestamp, self._function(*values)
            
        except (TypeError, IndexError, ZeroDivisionError) as err:
            # In multithreaded environments, there's a rare race condition
            # in which the main channel can be accessed before the calibration
            # channel has loaded. This should fix it.
            if getattr(self.dataset, 'loading', False):
                logger.warning("%s occurred in combined polynomial %r" %
                               (err.__class__.__name__, self))
                return None
            raise


    def inplace(self, values, y=None, timestamp=None, session=None, noBivariates=False, out=None):
        """ In-place transform for the `PolyPoly` transform.  It reduces the
            number of array allocations/operations compared to the normal
            `__call__` method.  The user can supply an `out` argument to save
            the results to that array, or leave `out` equal to `None`.  If `out`
            is `None`, then this method will allocate an array in the shape of
            `values`.

            This method does not do any calculations, and instead leaves that to
            its member polynomials.
        """
        if out is None:
            out = np.zeros_like(values, dtype=np.float64)

        try:
            # Optimization: don't check the other channel if Y is unused
            if self._noY is False:
                if noBivariates:
                    for i, poly in enumerate(self.polys):
                        poly.inplace(values[i], y=0, out=out[i])
                    return out

                session = self.dataset.lastSession if session is None else session
                sessionId = None if session is None else session.sessionId

                if self._eventlist is None or self._sessionId != sessionId:
                    channel = self.dataset.channels[self.channelId][self.subchannelId]
                    self._eventlist = channel.getSession(session.sessionId)
                    self._sessionId = session.sessionId

                # XXX: Hack! EventList length can be 0 if a thread is running.
                # This almost immediately gets fixed. Find real cause.
                try:
                    y = self._eventlist.getMean()
                except IndexError:
                    sleep(0.001)
                    if len(self._eventlist) == 0:
                        return None
                    y = self._eventlist.getMean()

                for i, poly in enumerate(self.polys):
                    poly.inplace(values[i], y=y, out=out[i])
                return out

            else:
                for i, poly in enumerate(self.polys):
                    poly.inplace(values[i], out=out[i])
                return out

        except (TypeError, IndexError, ZeroDivisionError) as err:
            # In multithreaded environments, there's a rare race condition
            # in which the main channel can be accessed before the calibration
            # channel has loaded. This should fix it.
            if getattr(self.dataset, 'loading', False):
                logger.warning("%s occurred in combined polynomial %r"%
                               (err.__class__.__name__, self))
                return None
            raise


    def isValid(self, session=None, noBivariates=False):
        """ Check the validity of the Transform.
        """
        if not Transform.isValid(self, session, noBivariates):
            return False
        return all(p.isValid(session, noBivariates) for p in self.polys)
